
import SwiftUI
import MetalKit


struct MetalView: View {
    @Binding var thetext: String
    
    @State private var metalView = MTKView()
    @State private var renderer: Renderer?
    
    // Set initial window size here - workaround for macOS 11 and 12
    // Need to add or subtract 28 for the menu bar height
    @State var minSize:CGSize = {
        @AppStorage("DefaultWindowWidth") var DefaultWindowWidth = 1000
        @AppStorage("DefaultWindowHeight") var DefaultWindowHeight = 500
        return CGSize(width: DefaultWindowWidth, height: DefaultWindowHeight)
    }()
    @State var maxSize:CGSize = {
        @AppStorage("DefaultWindowWidth") var DefaultWindowWidth = 1000
        @AppStorage("DefaultWindowHeight") var DefaultWindowHeight = 500
        return CGSize(width: DefaultWindowWidth, height: DefaultWindowHeight)
    }()
    
    // Sharing variables with the MetalView
    class SharedVarsStorage: ObservableObject {
        @Published var fps: String = ""
        @Published var feedback: String = ""
        @Published var info: String = ""
    }
    @StateObject var sharedVars = SharedVarsStorage()
    
    func updateViewVars(_ theVar: String, _ updatedText: String) {
        if theVar == "fps" { sharedVars.fps = updatedText }
        if theVar == "feedback" { sharedVars.feedback = updatedText }
        if theVar == "info" { sharedVars.info = updatedText }
    }
    
    struct FpsOverlay: View {
        @ObservedObject var sharedVars: SharedVarsStorage
        var body: some View {
            if sharedVars.fps != "" {
                Text(sharedVars.fps)
                    .font(.system(size: 16))
                    .padding(6)
                    .foregroundColor(.white)
                    .background(Color.gray)
                    .opacity(0.6)
                    .cornerRadius(5.0)
                    .padding(10)
            }
        }
    }
    
    struct FeedbackOverlay: View {
        @ObservedObject var sharedVars: SharedVarsStorage
        var body: some View {
            if sharedVars.feedback != "" {
                Text(sharedVars.feedback)
                    .font(.system(size: 16))
                    .padding(6)
                    .foregroundColor(.white)
                    .background(Color.gray)
                    .opacity(0.6)
                    .cornerRadius(5.0)
                    .padding(10)
            }
        }
    }
    
    struct HelpOverlay: View {
        @ObservedObject var sharedVars: SharedVarsStorage
        var body: some View {
            if sharedVars.info != "" {
                VStack(alignment: .leading) {
                    Text("Keyboard Actions")
                        .font(.system(size: 14) .bold())
                        .foregroundColor(.white)
                        .padding(2)
                    HStack {
                        // Info overlay text in two columns
                        Text(infoHelpKeys)
                            .font(.system(size: 14))
                            .foregroundColor(.white)
                            .padding(2)
                        Text(infoHelpText)
                            .font(.system(size: 14))
                            .foregroundColor(.white)
                            .padding(2)
                    }
                    .padding(0)
                }
                .padding(10)
                .foregroundColor(.white)
                .background(Color.gray)
                .opacity(0.6)
                .cornerRadius(5.0)
                .padding(10)
            }
        }
    }
    
    //init() { }
    
    
    // ###################################################################################################
    // VIEW BODY
    // ###################################################################################################
    
    var body: some View {
        MetalViewRepresentable(metalView: $metalView)
        
        // ###### Create renderer with the model ######
            .onAppear {
                renderer = Renderer(metalView: metalView, inputText: thetext, updateViewVars: updateViewVars)
            }
        
        // ###### If the file changes on disk then update the view ######
            .onChange(of: thetext) { newvalue in
                renderer!.readDataText(inputText: thetext, metalView: metalView)
            }
        
        
        // ###### Rotate object with mouse or trackpad ######
            .gesture(
                DragGesture(minimumDistance: 0.0, coordinateSpace: .local)
                    .onChanged { gesture in
                        let thepoint = CGPoint(x: gesture.translation.width, y: gesture.translation.height)
                        renderer!.dragUpdate(currentMousePoint: thepoint, dragEnded: false, metalView: metalView)
                    }
                    .onEnded { _ in
                        let thepoint = CGPoint(x: 0, y: 0)
                        renderer!.dragUpdate(currentMousePoint: thepoint, dragEnded: true, metalView: metalView)
                    }
            )
        
        
        // ###### Zoom Object ######
            .gesture (
                MagnificationGesture()
                    .onChanged { gesture in
                        let themag = gesture
                        renderer!.zoomUpdate(zoomChange: themag, zoomEnded: false, metalView: metalView)
                    }
                    .onEnded { gesture in
                        let themag = gesture
                        renderer!.zoomUpdate(zoomChange: themag, zoomEnded: true, metalView: metalView)
                    }
            )
        
        
        // ###### Overlays on screen ######
            .overlay(FpsOverlay(sharedVars: sharedVars), alignment: .bottomLeading)
            .overlay(FeedbackOverlay(sharedVars: sharedVars), alignment: .topLeading)
            .overlay(HelpOverlay(sharedVars: sharedVars), alignment: .topTrailing)
        
        
        // ###### Set initial window size to default ######
            .CUSTOMframe(minWidth: minSize.width, maxWidth: maxSize.width, minHeight: minSize.height+28, maxHeight: maxSize.height+28)
            .onAppear {
                // Fix up window resizing after window appears - workaround for macOS 11 and 12
                minSize = CGSize(width: 150, height: 150-28)
                maxSize = CGSize(width: 100000, height: 100000)
            }
        
        
        // ###### Scroll wheel and keyboard events - send to renderer ######
            .onAppear(perform: {
                NSEvent.addLocalMonitorForEvents(matching: [.keyDown, .scrollWheel]) { event in
                    var handled = false
                    
                    // Check if the event is relevant to this window
                    if event.windowNumber == metalView.window?.windowNumber {
                        
                        // Scroll wheel
                        if event.type == .scrollWheel {
                            renderer!.scrollUpdate(scrollX: event.scrollingDeltaX, scrollY: event.scrollingDeltaY, natScroll: event.isDirectionInvertedFromDevice, metalView: metalView)
                        }
                        
                        // Key down
                        if event.type == .keyDown {
                            // Remove device dependant modifier flags
                            let fixedModifier = event.modifierFlags.intersection(.deviceIndependentFlagsMask)
                            
                            // Pass all key codes and fixed modifer flags to renderer
                            handled = renderer!.keyUpdate(theKey: event.keyCode, theModifiers: fixedModifier.rawValue, metalView: metalView)
                            
                            // Toggle showing info as overlay (i key)
                            if event.keyCode == 34 && (fixedModifier.rawValue == 0 || fixedModifier.rawValue == 65536) {
                                if sharedVars.info == "" { sharedVars.info = "1"}
                                else { sharedVars.info = "" }
                                handled = true
                            }
                            
                            // Display keycode and modifier code as overlay - for identfying keys while developing
                            //sharedVars.feedback = String(event.keyCode) + " : " + String(fixedModifier.rawValue)
                        }
                    }
                    // If we handled the event then remove it from the queue otherwise pass it on
                    if handled { return nil } else { return event }
                }
            })
    }
}


struct MetalViewRepresentable: NSViewRepresentable {
    @Binding var metalView: MTKView
    func makeNSView(context: Context) -> some NSView { metalView }
    func updateNSView(_ uiView: NSViewType, context: Context) { }
}
